module Puff(
    Puff(..),
    Fragment(..),
    Signature(..),
    Key(..),
    FragmentList,
    HasFragmentList(..),
    Category, readPuffs, writePuffs, emptyPuff,
    catParseNew, catShowNew, parseCategoryOld, catShowOld, subCategory, fragmentData,
    getFragmentString, getFragmentStrings, showPuff, showFragments, 
    fragmentString, mergeFrags, getFragmentData, getFragmentTime
    ) where

import Word
import Int
import Time
import Bin
import ErrorLog
import List
import GenUtil
import SimpleParser

type Category = (String,String)

type FragmentList = [(String,Fragment)]

data Puff = Puff {cats :: [Category], signature :: [Signature], fragments :: FragmentList} 
data Fragment = FragmentText String | FragmentData [Word8] | FragmentTime ClockTime | FragmentInt Int32 | FragmentNest [(String,Fragment)]
data Signature = Unverifyable Key | Signed Key | Encrypted [String]
data Key = Key String [(String,Fragment)]

instance Eq Key where
    (==) (Key kn _) (Key kn' _) = kn == kn'

--------------------
-- Category Routines
--------------------

subCategory :: Category -> Category -> Bool
(x1,y1) `subCategory` (x2,y2) = x2 `isPrefixOf` x1 && y2 `isSuffixOf` y1

catfixes = [ ("/", ".|"), (".", "/"), (":", "..") ]

    
catParseNew :: String -> Category
catParseNew cs = (c,(drop 1) d) where
    (c,d) = span (/= '@') cs
    
parseCategoryOld :: Monad m => String -> m Category
parseCategoryOld = parser p where
    con cs | [nv] <- [x ++ (con $ drop (length y) cs) |(x,y) <- catfixes, y `isPrefixOf` cs] = nv
    con (c:cs) = c:con cs
    con "" = "" 
    bl [] = []
    bl [x] = []
    bl (x:xs) = x:bl xs
    p = do
	char '@'
	d <- many (noneOf "/")
	parseExact "/user/"
	c <- parseRest
	return (con (bl c),d)

catShowOld :: Category -> String
catShowOld (c,d) = "@" ++ d ++ "/user/" ++ con c ++ "/" where 
    con cs | [nv] <- [x ++ (con $ drop (length y) cs) |(y,x) <- catfixes, y `isPrefixOf` cs] = nv
    con (c:cs) = c:con cs
    con "" = "" 

catShowNew :: Category -> String
catShowNew (x,"") = x 
catShowNew (x,y) = x ++ "@" ++ y

---------------------
-- working with puffs
---------------------

emptyPuff = Puff {cats = [], signature = [], fragments = []}

fragmentData :: Monad m => String -> [(String,Fragment)] -> m [Word8]
fragmentData s fl = case lookup s fl of
    Just (FragmentData xs) -> return xs
    _ -> fail $ "fragment not found: " ++ s

fragmentString :: Monad m => String -> [(String,Fragment)] -> m String
fragmentString s fl = case lookup s fl of
    Just (FragmentText xs) -> return xs
    _ -> fail $ "fragment not found: " ++ s

mergeFrags :: [(String,Fragment)] -> [(String,Fragment)] -> [(String,Fragment)]
mergeFrags fla flb = fla ++ [f|f@(s,_) <- flb, s `notElem` fsts fla]


{-
getFragmentString :: Monad m => Puff -> String -> m String
getFragmentString (Puff {fragments = frags}) s = case lookup s frags of
    Just (FragmentText s) -> return s
    _ -> fail $ "fragment not found: " ++ s
-}

getFragmentStrings :: Puff -> String -> [String]
getFragmentStrings (Puff {fragments = frags}) s = [v|(n,FragmentText v) <- frags, n == s]

showPuff (Puff {cats = s, fragments =  fl}) = unwords (map catShowNew s) ++ " \n" ++ indentLines 2 (showFragments fl)
showFragments fl = concatMap (\(n,f) -> n ++ ": " ++ show f ++ "\n") fl

instance Show Puff where
    show = showPuff

instance Show Fragment where
    show (FragmentText s) = show s
    show (FragmentData xs) = "<DATA:" ++ show (length xs)  ++ ">"
    show (FragmentTime ct) = show ct
    show (FragmentInt x) = show x
    show (FragmentNest fl) = "Nest:" ++ (indentLines 2 $ showFragments fl)



getFragmentData :: (Monad m,HasFragmentList fl) => fl -> String -> m [Word8]
getFragmentData fl s = case [xs | (s',FragmentData xs) <- getFragmentList fl, s' == s] of
    (s:_) -> return s
    [] -> fail $ "data fragment not found: " ++ s

getFragmentString :: (Monad m, HasFragmentList fl)  => fl -> String -> m String
getFragmentString fl s = case [xs | (s',FragmentText xs) <- getFragmentList fl, s' == s] of
    (s:_) -> return s
    [] -> fail $ "text fragment not found: " ++ s

getFragmentTime :: (Monad m, HasFragmentList fl)  => fl -> String -> m ClockTime
getFragmentTime fl s = case [xs | (s',FragmentTime xs) <- getFragmentList fl, s' == s] of
    (s:_) -> return s
    [] -> fail $ "time fragment not found: " ++ s

class HasFragmentList a where
    getFragmentList :: a -> FragmentList

instance HasFragmentList FragmentList where
    getFragmentList = id

instance HasFragmentList Puff where
    getFragmentList (Puff {fragments = fl}) = fl

instance HasFragmentList (String,Fragment) where
    getFragmentList f = [f]

instance HasFragmentList Fragment where
    getFragmentList (FragmentNest fl) = fl
    getFragmentList _ = []

-- flattenFragmentList :: HasFragmentList fl => fl -> FragmentList

----------------
-- Puff Pickling
----------------

writePuffs :: String -> [Puff] -> IO ()
writePuffs fn ps = attempt $ putFile fn ps

readPuffs :: String -> IO [Puff]
readPuffs fn = tryElse [] $ getFile fn


---------------------
-- generated by DrIFT
---------------------

instance Binary Puff where
    put bh (Puff aa ab ac) = do
	    pos <- putBits bh 1 0
	    put bh aa
	    put bh ab
	    put bh ac
	    return pos
    get bh = do
	    h <- getBits bh 1
	    case h of
	      0 -> do
		    aa <- get bh
		    ab <- get bh
		    ac <- get bh
		    return (Puff aa ab ac)
    sizeOf (Puff aa ab ac) = 1 + sizeOf aa + sizeOf ab + sizeOf ac

instance Binary Fragment where
    put bh (FragmentText aa) = do
	    pos <- putBits bh 3 0
	    put bh aa
	    return pos
    put bh (FragmentData ab) = do
	    pos <- putBits bh 3 1
	    put bh ab
	    return pos
    put bh (FragmentTime ac) = do
	    pos <- putBits bh 3 2
	    put bh ac
	    return pos
    put bh (FragmentInt ad) = do
	    pos <- putBits bh 3 3
	    put bh ad
	    return pos
    put bh (FragmentNest ae) = do
	    pos <- putBits bh 3 4
	    put bh ae
	    return pos
    get bh = do
	    h <- getBits bh 3
	    case h of
	      0 -> do
		    aa <- get bh
		    return (FragmentText aa)
	      1 -> do
		    ab <- get bh
		    return (FragmentData ab)
	      2 -> do
		    ac <- get bh
		    return (FragmentTime ac)
	      3 -> do
		    ad <- get bh
		    return (FragmentInt ad)
	      4 -> do
		    ae <- get bh
		    return (FragmentNest ae)
    sizeOf (FragmentText aa) = 3 + sizeOf aa
    sizeOf (FragmentData ab) = 3 + sizeOf ab
    sizeOf (FragmentTime ac) = 3 + sizeOf ac
    sizeOf (FragmentInt ad) = 3 + sizeOf ad
    sizeOf (FragmentNest ae) = 3 + sizeOf ae

instance Binary Signature where
    put bh (Unverifyable aa) = do
	    pos <- putBits bh 2 0
	    put bh aa
	    return pos
    put bh (Signed ab) = do
	    pos <- putBits bh 2 1
	    put bh ab
	    return pos
    put bh (Encrypted ac) = do
	    pos <- putBits bh 2 2
	    put bh ac
	    return pos
    get bh = do
	    h <- getBits bh 2
	    case h of
	      0 -> do
		    aa <- get bh
		    return (Unverifyable aa)
	      1 -> do
		    ab <- get bh
		    return (Signed ab)
	      2 -> do
		    ac <- get bh
		    return (Encrypted ac)
    sizeOf (Unverifyable aa) = 2 + sizeOf aa
    sizeOf (Signed ab) = 2 + sizeOf ab
    sizeOf (Encrypted ac) = 2 + sizeOf ac

instance Binary Key where
    put bh (Key aa ab) = do
	    pos <- putBits bh 0 0
	    put bh aa
	    put bh ab
	    return pos
    get bh = do
	    h <- getBits bh 0
	    case h of
	      0 -> do
		    aa <- get bh
		    ab <- get bh
		    return (Key aa ab)
    sizeOf (Key aa ab) = 0 + sizeOf aa + sizeOf ab

--  Imported from other files :-
