module Signal(
    -- classes
    Changeable(..),
    Readable(..),
    -- Cache
    Cache,
    newCache,
    cacheInvalidate,
    cacheSignal,
    readCache,
    -- SVar
    SVar,
    newSVar,
    writeSVar,
    modifySVar,
    readSVar,
    svarSignal,
    -- signal
    Signal(..),
    newSignal,
    nonSignal,
    signalCondenser,
    -- watcher
    newWatcher,
    watcherHasChanged
    ) where

import Control.Monad.Trans
import Control.Concurrent
import Data.Unique

--------------------
-- Signals and SVars
--------------------

data Signal a = Signal {
    connect :: MonadIO m =>  (a -> IO ()) -> m (IO ()),
    disconnectAll :: MonadIO m => m (),
    signal :: MonadIO m => a -> m ()
    }


-- signals

signal_ :: Signal a -> Signal ()
signal_ s = s {connect = nc, signal = ns } where 
    nc f = connect s (\_ -> f ())
    ns _ = signal s undefined

newSignal :: MonadIO m => m (Signal a)
newSignal = do
    mv <- liftIO $ newMVar ([]::[(Unique,a -> IO ())])
    let c action = do
	    u <- liftIO newUnique
	    liftIO (modifyMVar_ mv (\as -> return ((u,action):as))) 
	    return (modifyMVar_ mv (\as -> return [ v|v@(u',_) <- as, u' /= u])) 
	d :: MonadIO m => m ()
	d = (liftIO $ swapMVar mv []) >> return ()
	s v =  do
	    cl <- liftIO $ readMVar mv
	    liftIO $ mapM_ (($ v) . snd) cl
    return $ Signal {connect = c, disconnectAll = d, signal = s} 


    
nonSignal = Signal {
    connect = \_ -> return (return ()),
    disconnectAll = return (),
    signal = \_ -> return ()
    }


signalCondenser :: [Signal a] -> (Signal a)
signalCondenser sl = nonSignal {
    connect = \f -> mapM (flip connect f) sl >>= \x -> return (sequence_ x),
    disconnectAll = mapM_ disconnectAll sl
    }


-- Cache
data Cache a = Cache {cacheValue :: !(MVar (Maybe a)), cacheFill :: !(IO a), cacheSignal :: !(Signal ())}

newCache :: IO a -> IO (Cache a)
newCache fill = do
    mv <- newMVar Nothing
    s <- newSignal
    return $ Cache {cacheValue = mv, cacheSignal = s, cacheFill = fill}

readCache :: Cache a -> IO a
readCache (Cache {cacheValue = mv, cacheFill = fill}) = modifyMVar mv $ \x -> case x of 
    v@(Just a) -> return (v,a)
    Nothing -> fill >>= \a -> return (Just a, a)

cacheInvalidate :: Cache a -> IO ()
cacheInvalidate c = swapMVar (cacheValue c) Nothing >> signal (cacheSignal c) () >> return ()

-- SVars

data SVar a = SVar (Signal a) (MVar a)

svarSignal :: SVar a -> Signal a
svarSignal (SVar s _) = s

newSVar :: MonadIO m => a -> m (SVar a)
newSVar v = do
    s <- newSignal 
    mv <- liftIO $ newMVar v
    return $ SVar s mv
writeSVar :: MonadIO m => (SVar a) -> a -> m ()
writeSVar (SVar s mv) v = do
    liftIO $ modifyMVar_ mv (const (return v))
    signal s v
modifySVar :: MonadIO m => (SVar a) -> (a -> a) -> m ()
modifySVar (SVar s mv) f = do
    v <- liftIO $ modifyMVar mv (\x -> (let y = f x in return (y,y)))
    signal s v
readSVar :: MonadIO m => (SVar a) -> m a
readSVar (SVar _ mv) = liftIO $ readMVar mv  

{-
modifySVar_ :: MonadIO m => (SVar a) -> (a -> IO a) -> m ()
modifySVar_ (SVar s mv) f = do
    v <- liftIO $ modifyMVar mv (\x -> (let y = f x in return (y,y)))
    signal s v
-}
    
data Watcher a = Watcher (MVar Bool) (IO a)

newWatcher :: forall c a. (Readable c, Changeable (c a)) => c a -> IO (Watcher a)
newWatcher x = do
    mv <- newMVar False
    connect (signalChanged x)  (\_ -> swapMVar mv True >> return ())
    return $ Watcher mv (readVal x)

instance Readable Watcher where
    readVal (Watcher mv x) = liftIO x

watcherHasChanged :: Watcher a -> IO Bool
watcherHasChanged (Watcher mv _) = readMVar mv
    

class Readable c where
    readVal :: MonadIO m => c a -> m a


class Changeable a where
    signalChanged :: a -> Signal ()


instance Changeable (Cache a) where
    signalChanged = cacheSignal

instance Changeable (SVar a) where
    signalChanged = signal_ . svarSignal

instance Readable Cache  where
    readVal x = liftIO $ readCache x

instance Readable SVar where
    readVal x = liftIO $ readSVar x


{-
class Frillable f where
    combine :: f a -> f b -> f (a,b)
    void :: f a -> f ()
    observe :: f a -> a -> IO b -> f b
-}
